/******************************************************************
 * Qiang Zheng,
 * ECS Dept, Syracuse Univ. Nov. 10, 1999
 * for CIS 615 class, Spring, 2000,
 * Instructor: Geoffrey C Fox, Nancy McCracken
 * O.S.: Solaris  Environment: VPL
 * Libraries: mpilib.a
 ******************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <mpi.h>

/*
 * TASK DESCRIPTION:
 * This program is to parallellize a sequential program which finds
 * the approximate value of the integral of the function
 * dx/sqrt( 1-( cos(x)*cos(x) )/2 ) over the interval 0 to PI/2
 * Following is the sequential program
 *
 * #include stdio.h
 * #include stdlib.h
 * #include math.h
 *
 * main()
 * {
 *  double deltax, pi = 3.141592654, integral, fleftx, frightx, 
 *         rightx, f();
 *  int nintervals = 128, i;
 *
 *  deltax = (0.5*pi)/(double)nintervals;
 *  integral = 0.0;
 *  fleftx = f(0.0);
 *  for (i=1; i<=nintervals; i++) 
 *   {
 *    rightx = deltax * (double) i;
 *    frightx = f(rightx);
 *    integral += deltax * 0.5 *(fleftx + frightx);
 *    fleftx = frightx;
 *   }
 *  printf ("\n Integral = %f\n", intergal);
 *  exit(0);
 * }
 * 
 * double f(double x)
 * {
 *  double y;
 *  
 *  y = cos(x);
 *  return (1.0/sqrt(1.0 - 0.5*y*y));
 * }
 */

#define PI	3.141592654
#define NUM_OF_INTERVALS	128
#define f(x)	( 1.0/sqrt( 1.0-0.5*( cos( x ) )*( cos( x ) ) ) )

main( int argc, char *argv[] )
{
    int myRank;
    double sTime, eTime;

    MPI_Init( &argc, &argv ); /* Initialize MPI environment */

    MPI_Comm_rank( MPI_COMM_WORLD, &myRank ); 

    sTime=MPI_Wtime();

    slave( myRank );          /* do the subpart computing job */

    if( myRank == 0 ){
        master();             /* find out the sum of the result from slave */
    }

    eTime=MPI_Wtime();

    if( myRank == 0 ) printf("Elapsed time= %f\n", eTime-sTime);

    MPI_Finalize();           /* cleanup MPI environment */

}


/*
 * The master function:
 * 1) partition the task into several parts 
 * 2) assign them to different slave processes,
 */
master( )
{
    int iterTimes, ntasks, rank;
    MPI_Status  status;
    double	result=0.0, sum=0.0;

    MPI_Comm_size( MPI_COMM_WORLD, &ntasks );

    printf ("\n the number of processes is: %d \n", ntasks);

    for( rank=0; rank<ntasks; rank++ ) {
        /* Prototype:
         * MPI_Recv( &ntasks, 1, MPI_INT, MPI_ANY_SOURCE
         *   , MPI_ANY_TAG, MPI_COMM_WORLD, &status );
         */
        MPI_Recv( &result, 1, MPI_DOUBLE, MPI_ANY_SOURCE
            , MPI_ANY_TAG, MPI_COMM_WORLD, &status );

        /* find the sum of the data collected from slaves */
        sum += result;
    }

    printf ( "\n The integral of \
                 f( x )= dx/sqrt( 1-( cos(x)*cos(x) )/2 ) \
                 over the interval 0 to PI/2 is %f \n", sum );
    return;
}

/*
 * the funtion of the slave:
 * 1) It finds out what to be computed according to its 
 *    rank and the number of the proces in this task
 * 2) It compute its part
 * 3) Finally, the slave returns the result
 * Notice: it is not necessary to exchange data with neighbors
 */ 
slave( int myRank )
{
    int         i, j, ntasks;
    MPI_Status  status;
    double		fLeftX, fRightX, leftX, rightX, subIntegral;
    double		deltax = ( 0.5*PI )/( double ) NUM_OF_INTERVALS;

    MPI_Comm_size( MPI_COMM_WORLD, &ntasks );

    leftX = myRank*( 0.5*PI )/( double )ntasks;

    fLeftX = f( leftX );
    for ( i=1; i<=NUM_OF_INTERVALS/ntasks; i++ ) {
      rightX = leftX+( double )i*deltax;
      fRightX = f( rightX );
      subIntegral += deltax*0.5*( fLeftX+fRightX );
      fLeftX = fRightX;
    }

    /* subIntegral = 0.5; */

    /* sent subtotal to master */
    /* 
     * prototype :
     * MPI_Send( buffer, count, datatype, destination, tag, MPI_COM_WORLD );
     */
    MPI_Send( &subIntegral, 1, MPI_DOUBLE, 0, 0, MPI_COMM_WORLD );

    printf("\n The subIntegral in process %d is %f \n", myRank, subIntegral);

    return;

/* free memory */
}
